#include <stdio.h>
#include <stdlib.h>
#include <FSextend.h>
#include "sqlifx.h"
#define BUFFERSIZE 100
EXEC SQL include sqlda;
EXEC SQL include varchar;
EXEC SQL include sqltypes;
short int year, month, day, RecordLen;
long nRows;
int HeaderLen=0, Version;

int write_in_file(FILE *tfp, char *sel_statement, short *nFields)
{
  EXEC SQL BEGIN DECLARE SECTION;
    struct db_Struct_File *MyRecords;
    char *retrieve_dir, *query;
    int i, j, pid, type_of_retrieve, verbosity, precision, scale;
    static int RelCount = 0;
    struct sqlda *db_ptr;
    struct sqlvar_struct *col_ptr;
    int cnt, size;
    long real_size;
    char format[300];
    char impre[200], linea[200];
    char outbuf[BUFFERSIZE];
  EXEC SQL END DECLARE SECTION;
  if((query=malloc(strlen(sel_statement)+1))==NULL) 
  {
    sprintf(errores,"%s", "I cannot get memory to put the query");
    pr_error(errores);
    return -1;
  }
  strcpy(query,sel_statement);

  err_log = getenv("MY_ERROR_LOG");
  verbosity = atoi(getenv("MY_LEVEL_VERBOSITY"));

  do
  {
    EXEC SQL prepare dbid from :query;
    if (sqlca.sqlcode < 0) break;
    EXEC SQL declare dbcursor cursor for dbid;
    if (sqlca.sqlcode < 0)
    {
        EXEC SQL free dbid;
      break;
    }
    EXEC SQL open dbcursor;
    if (sqlca.sqlcode < 0)
    {
        EXEC SQL free dbid;
        EXEC SQL free dbcursor;
      break;
    }
    EXEC SQL describe dbid into db_ptr;
    if (sqlca.sqlcode < 0)
    {
        EXEC SQL free dbid;
        EXEC SQL free dbcursor;
        free(db_ptr);
      break;
    }
  } while(0);
  if (sqlca.sqlcode < 0)
  {
    sprintf(errores,"%s code: %s", sqlca.sqlerrm,SQLSTATE);
    pr_error(errores);
    free(query);
    return -1;
  }

  *nFields = db_ptr->sqld;
  if((MyRecords = (struct db_Struct_File *) malloc( db_ptr->sqld * sizeof(struct db_Struct_File)))==NULL)
  {
    sprintf(errores,"%s", "I cannot get memory to put the xbase structure");
    pr_error(errores);
    /* Free all, because an error in malloc() */
    EXEC SQL close dbcursor;
    EXEC SQL free dbid; EXEC SQL free dbcursor; free(db_ptr); free(query);
    return -1;
  }

  /* We obtain the pointer to data for each elements returned in
     select */
  RecordLen=0;
  for(col_ptr=db_ptr->sqlvar, cnt=0; cnt < *nFields;
        cnt++, col_ptr++)
  {
    /* Sorry, but the length of column name in a dbf file it doesn't
       exceed 10 characters, so if it is longer than 10, I will 
       truncate the column name */

    if(strlen(col_ptr->sqlname)>10) 
      strncpy(MyRecords[cnt].FieldName,col_ptr->sqlname,10);
    else
      strcpy(MyRecords[cnt].FieldName,col_ptr->sqlname);
    MyRecords[cnt].FieldName[10]='\0';

      /* Allow for the trailing null character in C
           character arrays */
    switch(col_ptr->sqltype)
    {
      case SQLCHAR:
          col_ptr->sqllen += 1;
          break;
      case SQLVCHAR:
      case SQLNCHAR:
      case SQLNVCHAR:
          col_ptr->sqllen = VCMAX(col_ptr->sqllen)+1;
          break;
    }

      /* Determine size used by column data and increment
         buffer position */
      size = rtypmsize(col_ptr->sqltype, col_ptr->sqllen);
      size +=2;

    if((col_ptr->sqldata = malloc(size))==NULL)
    {
        sprintf(errores,"I cannot get memory for data, malloc() fail ");
        pr_error(errores);
      /* Due to an error allocating memory I must abandon all, but
         first free all the memory and cursor */
        for(i=0;i<cnt;i++)free(db_ptr->sqlvar[i].sqldata);
        EXEC SQL close dbcursor;
        EXEC SQL free dbid; EXEC SQL free dbcursor; free(db_ptr); free(query);
        free(MyRecords);
        return -1;
    }
   /* get the sql type of field and convert to xbase type */
    MyRecords[cnt].NoOfDecs=0;
    switch(col_ptr->sqltype)
    {
      case SQLSMINT:
      case SQLINT:
      case SQLSERIAL:
      	MyRecords[cnt].Type=DB_NUMERIC_FLD;
    	MyRecords[cnt].FieldLen=INT;
	break;
      case SQLFLOAT:
      case SQLSMFLOAT:
      	MyRecords[cnt].Type=DB_NUMERIC_FLD;
    	MyRecords[cnt].FieldLen=INT;
    	MyRecords[cnt].NoOfDecs=2;
	break;
      case SQLDECIMAL:
      case SQLMONEY:
	precision = (col_ptr->sqllen >> 8) & 0xFF;
	scale = (col_ptr->sqllen & 0xFF);
      	MyRecords[cnt].Type=DB_NUMERIC_FLD;
    	MyRecords[cnt].FieldLen=precision;
    	MyRecords[cnt].NoOfDecs=scale;
	break;
      case SQLDTIME:
      	MyRecords[cnt].Type=DB_CHAR_FLD;
    	MyRecords[cnt].FieldLen=TU_LEN(col_ptr->sqllen)+15;
	break;
      case SQLINTERVAL:
      	MyRecords[cnt].Type=DB_CHAR_FLD;
    	MyRecords[cnt].FieldLen=TU_LEN(col_ptr->sqllen)+15;
	break;
      case SQLCHAR:
      case SQLBYTES:
      case SQLTEXT:
      	MyRecords[cnt].Type=DB_CHAR_FLD;
	if(col_ptr->sqllen > 255)
    	  MyRecords[cnt].FieldLen=255;
	else
    	  MyRecords[cnt].FieldLen=col_ptr->sqllen-1;
	break;
      case SQLVCHAR:
      case SQLNCHAR:
      case SQLNVCHAR:
      	MyRecords[cnt].Type=DB_CHAR_FLD;
	if(col_ptr->sqllen > 255)
    	  MyRecords[cnt].FieldLen=255;
	else
    	  MyRecords[cnt].FieldLen=col_ptr->sqllen-1;
	break;
      case SQLDATE:
	MyRecords[cnt].FieldLen=8;
      	MyRecords[cnt].Type=DB_DATE_FLD;
	break;

  /* The next type of field are not supported yet, if you really
     want to use it, please let me know, because I am thinking in
     supporting it */
      default:
        sprintf(errores,"Code: %d, This type of data is not supported yet",
                 col_ptr->sqltype);
	pr_error(errores);
        for(i=0;i<*nFields;i++) free(db_ptr->sqlvar[i].sqldata);
        EXEC SQL close dbcursor;
        EXEC SQL free dbid; EXEC SQL free dbcursor; free(db_ptr); free(query);
        free(MyRecords);
        return -1;
	break;
    }
  }
  if(CreateStruct( tfp, MyRecords, *nFields,nRows )== DB_ERROR) 
  {
    for(i=0;i<*nFields;i++) free(db_ptr->sqlvar[i].sqldata);
    EXEC SQL close dbcursor;
    EXEC SQL free dbid; EXEC SQL free dbcursor; free(db_ptr); free(query);
    free(MyRecords);
    return DB_ERROR;
  }
  nRows=0;
  for (nRows=0;;nRows++)
  {
    long fecha_conve;
    linea[0]='\0';
    EXEC SQL fetch dbcursor using descriptor db_ptr;
    if (strncmp(SQLSTATE, "02", 2) == 0) /* end of tuples in select */
    {
      free(query);
      break;
    }
    if (sqlca.sqlcode < 0)
    {
      sprintf(errores,"%s code: %s", sqlca.sqlerrm,SQLSTATE);
      pr_error(errores);
      free(query);
      break;
    }

	/* Print out the returned values */
      
      for (i=0; i<*nFields; i++)
      {
        col_ptr=&(db_ptr->sqlvar[i]);
        format[0]='\0';
        switch(col_ptr->sqltype)
        {
          case SQLSMINT:
          case SQLINT:
          case SQLSERIAL:
            sprintf(format,"%%%dd",MyRecords[i].FieldLen);
            fprintf(tfp,format,*col_ptr->sqldata);
            break;
          case SQLCHAR:
          case SQLBYTES:
          case SQLTEXT:
          case SQLVCHAR:
          case SQLNCHAR:
          case SQLNVCHAR:
            sprintf(format,"%%%ds",MyRecords[i].FieldLen);
            fprintf(tfp,format,col_ptr->sqldata);
            break;
          case SQLFLOAT:
          case SQLSMFLOAT:
            sprintf(format,"%%%d.%df",MyRecords[i].FieldLen,
                 MyRecords[i].NoOfDecs);
            fprintf(tfp,format,* ((double *) (col_ptr->sqldata)));
            break;
          case SQLDECIMAL:
          case SQLMONEY:
            sprintf(format,"%%%ds",MyRecords[i].FieldLen);
            dectoasc((dec_t *)col_ptr->sqldata, outbuf, MyRecords[i].FieldLen+1, 
		2);
            outbuf[MyRecords[i].FieldLen]='\0';
            fprintf(tfp,format,outbuf);
            break;
          case SQLDTIME:
            sprintf(format,"%%%ds",MyRecords[i].FieldLen);
    	    dttoasc((dtime_t *) col_ptr->sqldata,outbuf);
            fprintf(tfp,format,outbuf);
	    break;
          case SQLINTERVAL:
            sprintf(format,"%%%ds",MyRecords[i].FieldLen);
    	    intoasc((intrvl_t *) col_ptr->sqldata,outbuf);
            fprintf(tfp,format,outbuf);
	    break;
          case SQLDATE:
            sprintf(format,"%%%ds",MyRecords[i].FieldLen);
            rfmtdate(* (long *) col_ptr->sqldata,"yyyymmdd",outbuf);
          /*  rdatestr(* (long *)(db_ptr->sqlvar[i].sqldata),outbuf);*/
            fprintf(tfp,format,outbuf);
            break;
        }
      }
      fprintf(tfp,"%s"," "); /* the space to mark deleted record */
  }
  if(WriteHeader( tfp, 2 ) == DB_ERROR )
  {
      pr_error("I cannot write the number of rows");
      for(i=0;i<*nFields;i++) free(db_ptr->sqlvar[i].sqldata);
      EXEC SQL close dbcursor;
      EXEC SQL free dbid; EXEC SQL free dbcursor; free(db_ptr); free(query);
      free(MyRecords);
      return DB_ERROR;
  }
  /* We need to explicitly free data because it was allocated with 
   * malloc().
   */
  for(i=0;i<*nFields;i++)free(db_ptr->sqlvar[i].sqldata);

  EXEC SQL close dbcursor;

  EXEC SQL free dbid;
  EXEC SQL free dbcursor;

  /* Free memory assigned to sqlda pointer. */
  free(db_ptr);
  free(query);
  free(MyRecords);
  return nRows;

}
int CreateStruct( FILE *dbf_file, struct db_Struct_File * sf, 
    const short nFields, const int nRows )
{

   int    j, k, k2, NameLen, rc;
   int MemoPresent;
   int i = j = 0;
   long CurRec;
   char *RecBuf, *RecBuf2;
   struct db_Struct_Rec *Struct_Rec_Ptr;
   char EofChar[10], impre[200];

   strncpy(EofChar,"\x0D\0x1A",10);


   /* i get the length of the record */
   MemoPresent=0;
   /* MEMO is not implemented yet, maybe in the near future */
   for (i=0;i<nFields;i++) 
   {
     if(sf[i].Type== DB_MEMO_FLD) MemoPresent=1;
   }
   RecordLen=0;
   for(i=0;i<nFields;i++)RecordLen+=sf[i].FieldLen; 
   RecordLen++; /* the place to put the deleted record */

   if(( RecBuf = (char *) malloc( RecordLen )) == NULL ) {
      pr_error(DB_MEMORY_ERROR);
      return DB_ERROR;
   }

   if(( RecBuf2 = (char *) malloc( RecordLen )) == NULL )
   { 
      free( RecBuf );
      pr_error(DB_MEMORY_ERROR);
      return DB_ERROR;
   }

   /* BlankRecord(); */
   memset( RecBuf, 0x20, RecordLen );
   memset( RecBuf2, 0x20, RecordLen );

   /* set class variables */

   if( MemoPresent ) 
      Version = 0x83;
   else
      Version = 0x03;

   CurRec  = 0L;

   HeaderLen = 33 + nFields * 32;

   year = 0x00;
   month = 0x02;
   day = 0x01;



   /* write the header information */
   if(( rc = WriteHeader( dbf_file, 0 )) != DB_NO_ERROR )
   {  
      free( RecBuf );
      free( RecBuf2 );
      pr_error(DB_ERROR_WRITING_HEADER);
      return DB_ERROR;
   }

   for( i = 0; i < 20; i++ )
   {
      if(( fwrite( "\x00", 1, 1, dbf_file )) != 1 )
      {
         free( RecBuf );
         free( RecBuf2 );
         pr_error(DB_ERROR_WRITING_HEADER);
         return DB_ERROR;
      }
   }
   if((Struct_Rec_Ptr = (struct db_Struct_Rec *) malloc( nFields * sizeof(struct db_Struct_Rec))) == NULL ) 
   {
      free( RecBuf ); 
      free( RecBuf2 );
      pr_error(DB_MEMORY_ERROR);
      return DB_ERROR;
   }
   memset( Struct_Rec_Ptr, 0, ( nFields * sizeof(struct db_Struct_Rec)));

   /* write the field information into the header */
   for( i = 0, k = 1; i < nFields; i++ )
   {
      strcpy( Struct_Rec_Ptr[i].FieldName, sf[i].FieldName );
      Struct_Rec_Ptr[i].Type = sf[i].Type;

      if( sf[i].Type == 'M' || sf[i].Type == 'B' || sf[i].Type == 'O' )
      {  /* memo fields are always 10 bytes */
        Struct_Rec_Ptr[i].FieldLen = 10;
        Struct_Rec_Ptr[i].NoOfDecs = 0;
      }
      else
      {
        Struct_Rec_Ptr[i].FieldLen = sf[i].FieldLen;
        Struct_Rec_Ptr[i].NoOfDecs = sf[i].NoOfDecs;
      }

      if( Struct_Rec_Ptr[i].NoOfDecs > Struct_Rec_Ptr[i].FieldLen )
      {
         free( Struct_Rec_Ptr );
         free( RecBuf );
         free( RecBuf2 );
         pr_error("Invalid schema");
         return DB_ERROR;
      }

      k2 = k;
      k += Struct_Rec_Ptr[i].FieldLen;

      if(( fwrite( &Struct_Rec_Ptr[i], 1, 18, dbf_file )) != 18 )
      {
         free( Struct_Rec_Ptr );
         free( RecBuf );
         free( RecBuf2 );
         pr_error(DB_ERROR_WRITING_HEADER);
         return DB_ERROR;
      }

      for( j = 0; j < 14; j++ )
      {
         if(( fwrite( "\x00", 1, 1, dbf_file )) != 1 )
         {
            free( Struct_Rec_Ptr );
            free( RecBuf );
            free( RecBuf2 );
            pr_error(DB_ERROR_WRITING_HEADER);
            return DB_ERROR;
         }
      }
      Struct_Rec_Ptr[i].Address  = RecBuf  + k2;
      Struct_Rec_Ptr[i].Address2 = RecBuf2 + k2;
   }

   /* write the end of header marker */
   if(( fwrite( EofChar, 2, 1, dbf_file )) != 1 )
   {
      free( Struct_Rec_Ptr );
      free( RecBuf );
      free( RecBuf2 );
      pr_error(DB_ERROR_WRITING_HEADER);
      return DB_ERROR;
   }
   if( MemoPresent )
   {
   /*   CreateMemoFile();*/
   }
}



int WriteHeader( FILE *dbf_file,const int PositionOption )
{
   char buf[4];
   if (PositionOption)
     rewind( dbf_file );

   /* write the version number */
   if(fwrite(&Version, 1, 1, dbf_file) != 1)
   {
     pr_error(DB_ERROR_WRITING_HEADER);
     return DB_ERROR;
   }
   
   if(fwrite(&year, 1, 1, dbf_file) != 1)
   {
     pr_error(DB_ERROR_WRITING_HEADER);
     return DB_ERROR;
   }

   if(fwrite(&month, 1, 1, dbf_file) != 1)
   {
     pr_error(DB_ERROR_WRITING_HEADER);
     return DB_ERROR;
   }

   if(fwrite(&day, 1, 1, dbf_file) != 1)
   {
     pr_error(DB_ERROR_WRITING_HEADER);
     return DB_ERROR;
   }

   memset( buf, 0x00, 4 );
   PutLong( buf, nRows );
   if (fwrite(&buf, 4, 1, dbf_file) != 1)
   {
     pr_error(DB_ERROR_WRITING_HEADER);
     return DB_ERROR;
   }

   if (PositionOption == 2)
      return DB_NO_ERROR;

   memset( buf, 0x00, 4 );
   PutUShort( buf, HeaderLen );
   if (fwrite(&buf, 2, 1, dbf_file) != 1)
   {
     pr_error(DB_ERROR_WRITING_HEADER);
     return DB_ERROR;
   }

   memset( buf, 0x00, 4 );
   PutUShort( buf, RecordLen );
   if (fwrite(&buf, 2, 1, dbf_file) != 1)
   {
     pr_error(DB_ERROR_WRITING_HEADER);
     return DB_ERROR;
   }

   return DB_NO_ERROR;
}
/* This routine puts a long value to a 4 byte character stream */
void PutLong( char * c, const long l )
{
   char *sp, *tp;
   int i;

   tp = c;
   sp = (char *) &l;
   if( EndianType == 'L' )
      for( i = 0; i < 4; i++ ) *tp++ = *sp++;
   else
   {
      sp+=3;
      for( i = 0; i < 4; i++ ) *tp++ = *sp--;
   }
   return;
}

/* This routine puts a short value to a 2 byte character stream */
void PutUShort( char * c, const short int s )
{
   char *sp, *tp;
   int i;

   tp = c;
   sp = (char *) &s;
   if( EndianType == 'L' )
      for( i = 0; i < 2; i++ ) *tp++ = *sp++;
   else
   {
      sp++;
      for( i = 0; i < 2; i++ ) *tp++ = *sp--;
   }
   return;
}
